/*
* https://www.eclipse.org/paho/files/jsdoc/Paho.MQTT.Client.html
* mqtt模块， 用于消息发布和订阅
* mqtt消息协议：
    压缩方式：（0：不压缩， 1：lz4压缩， 2: zstd压缩）-- 占2位
    是否差异比较：（0：否， 1：是）-- 占1位
    版本号：-- 占5位，------------------------以上总共1字节
    原始数据大小：仅在发布数据为压缩数据时需要，  类型为PInt, PInt类型参考./util/bon.ts
    剩余部分：消息内容
*/
import { baseType } from './lang/type';
import { BonBuffer } from './serialization/bon';
import { RSync } from './util/rsync';
import { Message, Client as MqttClientRaw } from '/@/utils/mqtt/mqtt';

export interface Connect {
    /**
     * 是否连接成功
     */
    connected: boolean;
    /**
     * 连接状态
     */
    connectState: ConnectState;
    /**
     * 建立连接
     */
    connect(connectOptions: any): void;
    /**
     * 重连
     */
    reconnect(): void;

    /**
     * 断开连接
     */
    disconnect(): void;

    /**
     * 连接丢失回调
     */
    onConnectionLost: (f: (info: DisConnectOption) => void) => void;

    /**
     * 连接成功回调
     */
    onConnected: (f: (reconnect: boolean, uri: string) => void) => void;

    /**
     * 网络状态改变
     */
    onConnectChange: (f: (state: ConnectState) => void) => void;
}

/**
 * mqtt客户端
 */
export interface MqttClient extends Connect {
    /**
     * 如果设置为true，则在断开与服务器的连接时将启用断开连接的发布。
     */
    disconnectedPublishing: boolean;
    /**
     * 用于设置拒绝缓冲区在拒绝新消息之前将保留的最大消息数。默认大小：5000条消息
     */
    disconnectedBufferSize: number;

    /**
     * 收到消息回调
     */
    onMessageArrived: (any);

    clientId: string;

    /**
     * 发布消息
     * @param topic 主题
     * @param payload 消息内容
     * @param qos 用于传递消息的服务质量。0：尽力而为（默认）； 1：至少一次； 2：恰好一次
     * @param retained 如果为true，则消息将由服务器保留并传递给当前和将来的订阅。如果为false，则服务器仅将消息传递给当前订阅者，这是新消息的默认设置。如果已发布消息且保留布尔值设置为true，并且在消息发布后进行了订阅，则收到的消息会将保留布尔值设置为true
     */
    publish(topic: string, payload: String | Uint8Array, qos?: number, retained?: boolean): void;

    /**
     * 监听指定主题的topic
     */
    onMessageWithTopic: (topic: string, cb: (payload: Uint8Array) => void) => void;

    /**
     * 注册消息监听器
     * @param cb 当收到消息时需要调用的回调方法，可重复注册多个回调
     */
    onMessage(cb: (topic: string, payload: Uint8Array) => void): void;

    /**
     * 订阅消息
     * @param filter 订阅主题 描述接收消息的目的地
     * @param subscribeOptions 用于控制订阅
     */
    subscribe(filter: string, subscribeOptions?: SubscribeOptions): void;
    /**
     * 取消订阅消息，停止接收发送到过滤器描述的目的地的消息
     * @param filter 描述接收消息的目的地
     * @param unsubscribeOptions 用于控制取消订阅
     */
    unsubscribe(filter: string, unsubscribeOptions: UnsubscribeOptions): void;

    /**
     * 网络速度监听
     */
    onSpeedChange: (f: (speed: number/**单位：ms*/) => void) => void;

    /**
     * 删除监听器
     * @param listenerType 监听器类型
     * @param f 监听函数
     */
    deleteListener: (listenerType: ListenerType, f: Function) => boolean;
}

export enum ConnectState {
    Connected, // 正常连接
    DisConnnected, // 连接丢失
    ReConnecting,	// 重连中
    Connecting, // 连接建立中
}

/**
 * 连接参数选项
 */
export interface ConnectOptions {
    timeout?: number;
    userName?: string;
    password?: string;
    willMessage?: Message;
    keepAliveInterval?: number;
    speedCheckInterval?: number;
    cleanSession?: boolean;
    useSSL?: boolean;
    invocationContext?: Object;
    onSuccess?: Function;
    onFailure?: Function;
    hosts?: string[];
    ports?: number[];
    reconnect?: boolean;
    mqttVersion?: number;
    mqttVersionExplicit?: boolean;
    uris?: string[];
}

/**
 * 用于控制订阅
 */
export interface SubscribeOptions {
    /**
     * 订阅后发送的所有出版物的最高质量要求
     */
    qos?: number,
    /**
     * 传递给onSuccess回调或onFailure回调
     */
    invocationContext?: any;
    /**
     * 从服务器收到订阅确认时调用。单个响应对象参数传递到包含以下字段的onSuccess回调：
     *	1.invocationContext（如果在subscriptionOptions中设置）。
     */
    onSuccess?: Function;
    /**
     * 订阅请求失败或超时时调用。单个响应对象参数传递到包含以下字段的onFailure回调:
     * 	1.invocationContext-如果在subscriptionOptions中设置。
     * 	2.errorCode-一个数字，指示错误的性质。
     * 	3.errorMessage-描述错误的文本。
     */
    onFailure?: Function;

    /**
     * 如果存在，则确定调用onFailure回调的秒数。超时的存在不会阻止订阅完成时调用onSuccess回调。
     */
    timeout?: number;
}

/**
 * 用于控制订阅
 */
export interface UnsubscribeOptions {
    /**
     * 传递给onSuccess回调或onFailure回调
     */
    invocationContext?: any;
    /**
     * 从服务器收到取消订阅确认时调用。单个响应对象参数传递到包含以下字段的onSuccess回调：
     *	1.invocationContext（如果在unsubscriptionOptions中设置）。
     */
    onSuccess?: Function;
    /**
     * 取消请求失败或超时时调用。单个响应对象参数传递到包含以下字段的onFailure回调:
     * 	1.invocationContext-如果在unsubscriptionOptions中设置。
     * 	2.errorCode-一个数字，指示错误的性质。
     * 	3.errorMessage-描述错误的文本。
     */
    onFailure?: Function;

    /**
     * 如果存在，则确定调用onFailure回调的秒数。超时的存在不会阻止取消订阅完成时调用onSuccess回调。
     */
    timeout?: number;
}

/**
 * @param {boolean} reconnect-如果为true，则表明连接是重新连接的结果
 * @param {string} url-用于连接到服务器的URI
 */
export interface OnConnectCall {
    (reconnect: boolean, url: string)
}

export interface DisConnectOption {
    errorCode: number,
    errorMessage: string,
    reconnect: boolean,
    uri: string,
};

export enum ListenerType {
    Connected = "connectedListeners",
    Disconnected = "disconnectedListeners",
    SpeedChange = "speedChangeListeners",
    ConnectChange = "connectChangeListeners",
}

/**
 * mqtt客户端
 * @example
 */
export class Client implements MqttClient {
    public mc: MqttClientRaw;
    public _connectState: ConnectState;
    public lastMsg = new Map<string, Uint8Array>();// 需要使用差异比较进行数据同步的主题，每次发布时，应该保存最后发布的数据
    // 可以为主题设置0个或多个tag，其中"compressMode"，"isRsync"为内置tag，用于配置主题数据的压缩模式，是否使用差异比较进行数据同步
    public tags = new Map<string, Map<string, baseType>>();
    public rsync: RSync;
    public compressTap: number;// 压缩阀值， 当数据大小超过compressTap， 将会使用tag中配置的压缩方式进行压缩
    private listeners = [];
    private topicListeners: Map<string, ((payload: Uint8Array) => void)[]> = new Map();
    private disconnectedListeners = [];
    private connectedListeners = [];
    private connectChangeListeners = [];
    private speedChangeListeners = [];
    constructor(host: string, port: number, clientId: string, compressTap?: number, connectOptions?: ConnectOptions) {
        const c = new MqttClientRaw(host, port, clientId);
        const scope = this;
        const onmessage = (message: any) => {
            scope.onMessageArrived(message);
        };
        c.onMessageArrived = onmessage;
        c.onConnectionLost = (info: any) => {
            forAndAdjust(this.disconnectedListeners, (l) => {
                l(info)
            })
            let state: ConnectState;
            if (info && info.reconnect) {
                state = ConnectState.ReConnecting;
            } else {
                state = ConnectState.DisConnnected;
            }
            this.notifyStateChange(state);
        };
        c.onConnected = (reconnect: boolean, uri: string) => {
            forAndAdjust(this.connectedListeners, (l) => {
                l(reconnect, uri);
            });
            this.notifyStateChange(ConnectState.Connected);
        };
        c.onSpeedChange = (speed: number) => {
            forAndAdjust(this.speedChangeListeners, (l) => {
                l(speed);
            });
        }
        if (connectOptions) {
            c.connect(connectOptions);
            this._connectState = ConnectState.Connecting;
        }
        this.mc = c;
        this.rsync = new RSync(32);
        this.compressTap = compressTap || 64;
    }

    public set disconnectedPublishing(value: boolean) {
        this.mc.disconnectedPublishing = value;
    }

    public set disconnectedBufferSize(value: number) {
        this.mc.disconnectedBufferSize = value;
    }

    public get connected(): boolean {
        return this.mc.isConnected();
    }

    public get trace() {
        return this.mc.trace;
    }
    public set trace(newTraceFunction: Function) {
        this.trace(newTraceFunction);
    }

    public get clientId(): string {
        return this.mc.clientId;
    }

    public get connectState(): ConnectState {
        return this._connectState;
    }

    // 重连
    public reconnect() {
        this.mc.reconnect();
        if (this._connectState !== ConnectState.ReConnecting) {
            this.notifyStateChange(ConnectState.ReConnecting);
        }
    }

    /** 
     * @param {number} connectOptions.timeout - 如果连接在该秒数内未成功，则视为失败了，默认值为30秒
     * @param {string} connectOptions.userName - 此连接的身份验证用户名.
     * @param {string} connectOptions.password - 此连接的身份验证密码.
     * @param {Paho.MQTT.Message} connectOptions.willMessage - 客户端异常断开时由服务器发送.
     * @param {number} connectOptions.keepAliveInterval - 如果在该秒数内没有活动，服务器将断开此客户端的连接.
     *                    如果未设置，则假定默认值为60秒.
     *  @param {number} connectOptions.speedCheckInterval - 检查网络速度的秒数间隔，默认等于keepAliveInterval.
     * @param {boolean} connectOptions.cleanSession - 如果为true（默认值），则在成功连接时将删除客户端和服务器的持久状态.
     * @param {boolean} connectOptions.useSSL - 如果存在且为true，则使用SSL Websocket连接.
     * @param {object} connectOptions.invocationContext - 传递给onSuccess回调或onFailure回调.
     * @param {function} connectOptions.onSuccess - 从服务器接收到连接确认时调用.
     * 当从服务器接收到连接确认时调用单个响应对象参数传递给包含以下字段的onSuccess回调:
     * <ol>
     * <li>传入connectOptions中onSuccess方法的调用上下文.
     * </ol>
     * @param {function} connectOptions.onFailure - 当连接请求失败或超时时调用.
     * 一个响应对象参数被传递到包含以下字段的onFailure回调:
     * <ol>
     * <li>传入connectOptions中onFailure方法的调用上下文.
     * <li>错误代码指示错误性质的数字.
     * <li>错误的文本描述.
     * </ol>
     * @param {array} connectOptions.hosts - 如果存在，则包含一组主机名或完全限定的主机名
     * WebSocket URI（ws://iot.eclipse.org:80/ws)，按顺序进行了测试
     * 构造器上的主机和端口参数。主机一次只测试一个，直到其中一个成功了
     * @param {array} connectOptions.ports - 如果存在与主机匹配的端口集。如果主机包含uri，则不使用此属性.
     * @param {boolean} connectOptions.reconnect - 设置在连接丢失时客户端是否自动尝试重新连接到服务器.
     *<ul>
    *<li>如果设置为false，则在连接丢失的情况下，客户端不会尝试自动重新连接到服务器.</li>
    *<li>如果设置为true，则在连接丢失的情况下，客户端将尝试重新连接到服务器，它将在尝试重新连接之前先等待1秒钟，
    * 对于每次失败的重新连接尝试，延迟将加倍，直到达到2分钟，此时延迟将保持在2分钟.</li>
    *</ul>
    * @param {number} connectOptions.mqttVersion - 用于连接到MQTT代理的MQTT版本.
    *<ul>
    *<li>3 - MQTT V3.1</li>
    *<li>4 - MQTT V3.1.1</li>
    *</ul>
    * @param {boolean} connectOptions.mqttVersionExplicit - 如果设置为true，将强制连接使用选定的MQTT版本，否则将无法连接.
    * @param {array} connectOptions.uris - 如果存在，则应包含完全限定的WebSocket uri列表（例如ws://iot.eclipse.org:80/ws)，
    * 以代替construtor的host和port参数， 每次尝试一个uri，直到其中一个成功。
    * 不要将其与主机一起使用，因为主机阵列将转换为uri并将覆盖此属性.
    * @throws {InvalidState} 如果客户端未处于断开连接状态。在第二次或以后的时间调用connect之前，客户端必须已接收到connectionLost或
    * disconnected.
    */
    public connect(connectOptions: ConnectOptions) {
        this.mc.connect(connectOptions);
        if (this._connectState !== ConnectState.Connecting) {
            this.notifyStateChange(ConnectState.Connecting);
        }
    }

    // 设置tag
    public setTag(topic: string, key: string, value: baseType) {
        let tag = this.tags.get(topic);
        if (!tag) {
            tag = new Map<string, baseType>();
            this.tags.set(topic, tag);
        }
        tag.set(key, value);
    }

    // 设置压缩阀值
    public setCompressTap(value: number) {
        this.compressTap = value;
    }

    /**
     * @description 发布消息
     */
    public publish(topic: string, message: Uint8Array, qos = 0, retained = false) {
        const messageHeadBb = new BonBuffer();
        let tag = this.tags.get(topic), isRsync, compressMode, originalSize;
        if (tag) {
            isRsync = tag.get('isRsync');
            compressMode = tag.get('compressMode');
        } else {
            isRsync = 0;// 默认不进行差异比较
            compressMode = 1;// 默认lz4压缩
        }

        // 如果需要进行差异比较，将发布数据改为差异数据
        // if(isRsync){
        // 	let last = this.lastMsg.get(topic);
        // 	this.lastMsg.set(topic, message);
        // 	if(!last){
        // 		isRsync = 0;
        // 	}else{
        // 		let bb1 = new BonBuffer()
        // 		encodeDiffs(this.rsync.diff(message, this.rsync.checksum(last)), bb1);
        // 		message = bb1.getBuffer();
        // 	}	
        // }

        // //如果数据大于压缩阀值，对其进行压缩 
        // if(message.length > this.compressTap){
        // 	originalSize = message.length;
        // 	if(compressMode === CompressMode.LZ4){
        // 		message = lz4.compress(message);
        // 	}else if(compressMode === CompressMode.ZSTD){
        // 		//todo
        // 	}else if(compressMode === CompressMode.NONE){
        // 		compressMode = 0;
        // 	}else{
        // 		throw "压缩方式不支持， mode：" + compressMode;
        // 	}	
        // }else{
        compressMode = 0;
        // }

        const first = (isRsync << 2) + compressMode;
        messageHeadBb.writeU8(first);
        originalSize && messageHeadBb.writePInt(originalSize);

        const messageHead = messageHeadBb.getBuffer();
        const u8 = new Uint8Array(messageHead.length + message.length);
        u8.set(messageHead);
        u8.set(message, messageHead.length);
        this.mc.publish(topic, u8, qos, retained);
    }

    /**
     * @description 订阅消息
     */
    public subscribe(filter: string, subscribeOptions?: SubscribeOptions) {
        if (!subscribeOptions) {
            subscribeOptions = { qos: 0 };
        } else if (!subscribeOptions.qos) {
            subscribeOptions.qos = 0;
        }
        this.mc.subscribe(filter, subscribeOptions);
    }

    /**
     * @description 退订
     */
    public unsubscribe(filter: string, unsubscribeOptions?: SubscribeOptions) {
        this.mc.unsubscribe(filter, unsubscribeOptions);
    }

    /**
     * @description 关闭连接
     */
    public disconnect() {
        try {
            this.mc.disconnect();
        } catch (error) {
            console.warn(error);
        }
    }

    public onConnectionLost(f: (param: DisConnectOption) => void) {
        this.disconnectedListeners.push(f);
    }

    public onConnected(f: (reconnect?: boolean, uri?: string) => void) {
        this.connectedListeners.push(f);
        if (this.connected) {
            f(false, this.mc._getURI());
        }
    }

    public onConnectChange(f: (state: ConnectState) => void) {
        this.connectChangeListeners.push(f);
    }

    public onSpeedChange(f: (speed: number) => void) {
        this.speedChangeListeners.push(f);
    }

    /**
     * 监听消息
     */
    public onMessage(cb: (topic: string, payload: Uint8Array) => void) {
        this.listeners.push(cb);
    }

    /**
    * 监听指定主题的消息
    */
    public onMessageWithTopic(topic: string, cb: (payload: Uint8Array) => void) {
        let topicListeners = this.topicListeners.get(topic);
        if (!topicListeners) {
            topicListeners = [];
            this.topicListeners.set(topic, topicListeners);
        }
        topicListeners.push(cb);
    }

    public deleteListener(listenerType: ListenerType, f: Function): boolean {
        const listeners = this[listenerType];
        if (listeners) {
            for (let i = 0, len = listeners.length; i < len; i++) {
                if (listeners[i] === f) {
                    delete listeners[i];
                    return true;
                }
            }
        }
        return false;
    }

    public onMessageArrived(message: Message) {
        try {
            let payload = message.payloadBytes;
            const c = (payload[0] >> 6 & 3);// 压缩方式（0：不压缩，1：lz4压缩, 2:zstd压缩）
            const r = (payload[0] >> 5 & 1);// 是否为差异部分
            payload = new Uint8Array(payload.buffer, payload.byteOffset + 1);
            if (c) {
                const view = new DataView(payload.buffer.slice(payload.byteOffset, payload.byteOffset + 4));
                const len = view.getUint32(0, true);
                payload = new Uint8Array(payload.buffer, payload.byteOffset + 4);
                if (c === 1) {
                    // payload = lz4.decompress(payload, len);
                } else if (c === 2) {
                    // todo
                } else {
                    throw new Error('压缩方式不支持，mode：' + c);
                }
            }

            const topic = message.destinationName;
            for (let i = 0; i < this.listeners.length; i++) {
                this.listeners[i](topic, payload);
            }
            // 通知监听当前主题的监听器
            let tl = this.topicListeners.get(topic);
            if (tl) {
                for (let l of tl) {
                    l(payload);
                }
            }
        } catch (error) {
            // mqtt会捕获并消耗掉该异常，并移除websokect的监听函数，导致无法收到后续消息却难以找到找到线索，因此此处将错误信息打印出来
            console.log(error);
            throw error;
        }
    }

    private notifyStateChange(state: ConnectState) {
        this._connectState = state;
        forAndAdjust(this.speedChangeListeners, (l) => {
            l(state);
        });
    }
}

export const forAndAdjust = (arr: any[], f: (item) => void) => {
    for (let i = 0; i < arr.length; i++) {
        const item = arr[i];
        if (item) {
            f(item);
        } else {
            arr.splice(i, 1);
            i--;
        }
    }
}
// // 压缩模式
// enum CompressMode {
//     NONE = 0,
//     LZ4 = 1,
//     ZSTD = 2
// }
