/**
 * RPC， 远程方法调用
 * 采用 mqtt上定义的每会话的$req和$resp主题，来发送请求和接受响应
 */

// ============================== 导入
import { Func } from "./serialization/struct_mgr";
import { BonBuffer, BonEncode } from './serialization/bon';
import { MqttClient } from "./mqtt_c";

export interface CallBack {
	(param: any ): void
} 

/**********************旧版本的rpc接口，rpc请求结果通过回调函数通知，会逐步废弃，不建议使用*********************************/
/**
 * 创建一个RPC函数
 * @example
 */
export const create = (client: MqttClient): Rpc => {
	const mqttRpc = createMqttRpcClient(client);
	return (name: string, req: BonEncode, callback: CallBack, timeout: number) => {
		mqttRpc.call(name, req, callback, timeout);
	}
}

export interface Rpc {
	(name: string, req: BonEncode, callback: CallBack, timeout: number): void;
}
/**********************旧版本的rpc接口，rpc请求结果通过回调函数通知，会逐步废弃，不建议使用*********************************/


/**************************************新版本的rpc接口，使用Promise异步返回结果*********************************************/
/**
 * 创建一个RPC函数
 * @example
 */
export const createRpc = (client: MqttClient): RequestRpc => {
	const mqttRpc = createMqttRpcClient(client);
	return (name: string, req: BonEncode, timeout: number) => {
		return new Promise((resolve, _reject/*处理异常，TODO*/) => {
			mqttRpc.call(name, req, resolve, timeout || 10000);
		});
	}
}

export interface RequestRpc {
	(name: string, req: BonEncode, timeout?: number): Promise<any>;
}
/**************************************新版本的rpc接口，使用Promise异步返回结果*********************************************/



/***************************************************本模块通用逻辑*************************************************/
export class MqttRpc {
	static rid = 1;
	wait = {};
	client: MqttClient;
	constructor(client: MqttClient) {
		this.client = client;
	};

	//远程调用
	// call(name: string, req: BonEncode, callback: Func, timeout: number) {
	// 	let bb = new BonBuffer();
	// 	this.wait[this.rid] = callback;
	// 	bb.writeU32(this.rid++); // rid
	// 	bb.writeU8(timeout); // timeout
	// 	this.rid >= 0xffffffff && (this.rid = 1);
	// 	bb.write(req);
	// 	this.client.publish(name, bb.getBuffer(), 0, true);
	// }

	//远程调用
	call(name: string, req: BonEncode, callback: Func, timeout: number): Uint8Array {
		let rid = MqttRpc.rid++;
		MqttRpc.rid >= 0xffffffff && (MqttRpc.rid = 1);
		let bb = this.serialize(req, timeout, rid);
		this.wait[rid] = callback;
		let buffer = bb.getBuffer()
		this.client.publish(name, buffer, 0, true);
		// 返回序列化后的数据
		return buffer;
	}

	/**
	 * 序列化rpc消息
	 */
	serialize(req: BonEncode, timeout: number, rid: number): BonBuffer {
		let bb = new BonBuffer();
		bb.writeU32(rid); // rid
		bb.writeU8(timeout); // timeout
		bb.write(req);
		return bb;
	}
}

export const createMqttRpcClient = (client: MqttClient): MqttRpc => {
	const mqttRpc = new MqttRpc(client);
	client.onMessageWithTopic("$r", (payload: Uint8Array) => {
		let bb = new BonBuffer(payload, 0, payload.length);
		let rid = bb.readU32();//消息开始表示此次请求的id
		bb.readU8(); // timeout TODO
		if (mqttRpc.wait[rid]) {
			mqttRpc.wait[rid](bb.read());
			delete mqttRpc.wait[rid];
		}
	});
	return mqttRpc;
}
