
/**
 * KTX 压缩纹理解析
 */
// export const ETC_COMPRESS_FLAG = '.etc.ktx';
// export const PVRTC_COMPRESS_FLAG = '.pvrtc.ktx';

export interface KTXInfo {
	mipmaps: {
        data: Uint8Array,
        width: number,
        height: number
    }[];
	width: number;
	height: number;
	format: number;
	isCubemap: boolean;
	mipmapCount: number;
}

export const KTX = { 
	parse: (buffer: ArrayBuffer | Uint8Array, loadMipmaps: boolean): KTXInfo => {

		// if (!(buffer instanceof ArrayBuffer)) {
		//     buffer = new ArrayBuffer();
		// }
		let ktx: KhronosTextureContainer;
		if(buffer instanceof ArrayBuffer) {
			ktx = new KhronosTextureContainer(buffer, 0, 1);
		} else {
			ktx = new KhronosTextureContainer(buffer.buffer, buffer.byteOffset, 1);
		}
	
		return {
			mipmaps: ktx.mipmaps(loadMipmaps),
			width: ktx.pixelWidth,
			height: ktx.pixelHeight,
			format: ktx.glInternalFormat,
			isCubemap: ktx.numberOfFaces === 6,
			mipmapCount: ktx.numberOfMipmapLevels
		};
	}
}

class KhronosTextureContainer {
    public static HEADER_LEN: number = (13 * 4) + 12; // identifier + header elements (not including key value meta-data pairs)
    public static COMPRESSED_2D: number = 0; // uses a gl.compressedTexImage2D()
    public static COMPRESSED_3D: number = 1; // uses a gl.compressedTexImage3D()
    public static TEX_2D: number = 2; // uses a gl.texImage2D()
    public static TEX_3D: number = 3; // uses a gl.texImage3D()
	public arrayBuffer: ArrayBuffer;
	public byteoffset: number;
    public glType: any;
    public glTypeSize: any;
    public glFormat: any;
    public glInternalFormat: any;
    public glBaseInternalFormat: any;
    public pixelWidth: any;
    public pixelHeight: any;
    public pixelDepth: any;
    public numberOfArrayElements: any;
    public numberOfFaces: any;
    public numberOfMipmapLevels: any;
    public bytesOfKeyValueData: any;
    public loadType: any;
    constructor(arrayBuffer: ArrayBuffer, byteoffset: number, facesExpected: number) {
        
		this.arrayBuffer = arrayBuffer;
		this.byteoffset = byteoffset;

        // Test that it is a ktx formatted file, based on the first 12 bytes, character representation is:
        // '´', 'K', 'T', 'X', ' ', '1', '1', 'ª', '\r', '\n', '\x1A', '\n'
        // 0xAB, 0x4B, 0x54, 0x58, 0x20, 0x31, 0x31, 0xBB, 0x0D, 0x0A, 0x1A, 0x0A
        const identifier = new Uint8Array(this.arrayBuffer, 0 + this.byteoffset, 12);
        if (identifier[ 0 ] !== 0xAB ||
            identifier[ 1 ] !== 0x4B ||
            identifier[ 2 ] !== 0x54 ||
            identifier[ 3 ] !== 0x58 ||
            identifier[ 4 ] !== 0x20 ||
            identifier[ 5 ] !== 0x31 ||
            identifier[ 6 ] !== 0x31 ||
            identifier[ 7 ] !== 0xBB ||
            identifier[ 8 ] !== 0x0D ||
            identifier[ 9 ] !== 0x0A ||
            identifier[ 10 ] !== 0x1A ||
            identifier[ 11 ] !== 0x0A) {

            console.error('texture missing KTX identifier');
            
            return;

        }

        // load the reset of the header in native 32 bit uint
        const dataSize = Uint32Array.BYTES_PER_ELEMENT;
        const headerDataView = new DataView(this.arrayBuffer, 12 + this.byteoffset, dataSize * 13);
        const endianness = headerDataView.getUint32(0, true);
        const littleEndian = endianness === 0x04030201;

        this.glType = headerDataView.getUint32(dataSize * 1, littleEndian); // must be 0 for compressed textures
        this.glTypeSize = headerDataView.getUint32(dataSize * 2, littleEndian); // must be 1 for compressed textures
        this.glFormat = headerDataView.getUint32(dataSize * 3, littleEndian); // must be 0 for compressed textures
        this.glInternalFormat = headerDataView.getUint32(dataSize * 4, littleEndian); // the value of arg passed to gl.compressedTexImage2D(,,x,,,,)
        this.glBaseInternalFormat = headerDataView.getUint32(dataSize * 5, littleEndian); // specify GL_RGB, GL_RGBA, GL_ALPHA, etc (un-compressed only)
        this.pixelWidth = headerDataView.getUint32(dataSize * 6, littleEndian); // level 0 value of arg passed to gl.compressedTexImage2D(,,,x,,,)
        this.pixelHeight = headerDataView.getUint32(dataSize * 7, littleEndian); // level 0 value of arg passed to gl.compressedTexImage2D(,,,,x,,)
        this.pixelDepth = headerDataView.getUint32(dataSize * 8, littleEndian); // level 0 value of arg passed to gl.compressedTexImage3D(,,,,,x,,)
        this.numberOfArrayElements = headerDataView.getUint32(dataSize * 9, littleEndian); // used for texture arrays
        this.numberOfFaces = headerDataView.getUint32(dataSize * 10, littleEndian); // used for cubemap textures, should either be 1 or 6
        this.numberOfMipmapLevels = headerDataView.getUint32(dataSize * 11, littleEndian); // number of levels; disregard possibility of 0 for compressed textures
        this.bytesOfKeyValueData = headerDataView.getUint32(dataSize * 12, littleEndian); // the amount of space after the header for meta-data

        // Make sure we have a compressed type.  Not only reduces work, but probably better to let dev know they are not compressing.
        if (this.glType !== 0) {

            console.warn('only compressed formats currently supported');
            
            return;

        } else {

            // value of zero is an indication to generate mipmaps @ runtime.  Not usually allowed for compressed, so disregard.
            this.numberOfMipmapLevels = Math.max(1, this.numberOfMipmapLevels);

        }
        if (this.pixelHeight === 0 || this.pixelDepth !== 0) {

            console.warn('only 2D textures currently supported');

            return;

        }
        if (this.numberOfArrayElements !== 0) {

            console.warn('texture arrays not currently supported');

            return;

        }
        if (this.numberOfFaces !== facesExpected) {

            // tslint:disable-next-line:prefer-template
            console.warn('number of faces expected' + facesExpected + ', but found ' + this.numberOfFaces);

            return;

        }
        // we now have a completely validated file, so could use existence of loadType as success
        // would need to make this more elaborate & adjust checks above to support more than one load type
        this.loadType = KhronosTextureContainer.COMPRESSED_2D;
    }
    public mipmaps = (loadMipmaps: boolean) => {

        const mipmaps = [];

        // initialize width & height for level 1
        let dataOffset = KhronosTextureContainer.HEADER_LEN + this.bytesOfKeyValueData + this.byteoffset;
        let width = this.pixelWidth;
        let height = this.pixelHeight;
        const mipmapCount = loadMipmaps ? this.numberOfMipmapLevels : 1;

        for (let level = 0; level < mipmapCount; level ++) {

			const imageSize = new DataView(this.arrayBuffer, dataOffset, 4).getInt32(0, true);
            // const imageSize = new Int32Array(this.arrayBuffer, dataOffset, 1)[ 0 ]; // size per face, since not supporting array cubemaps
            dataOffset += 4; // size of the image + 4 for the imageSize field

            for (let face = 0; face < this.numberOfFaces; face ++) {

                const byteArray = new Uint8Array(this.arrayBuffer, dataOffset, imageSize);

                mipmaps.push({ data: byteArray, width: width, height: height });

                dataOffset += imageSize;
                dataOffset += 3 - ((imageSize + 3) % 4); // add padding for odd sized image

            }
            width = Math.max(1.0, width * 0.5);
            height = Math.max(1.0, height * 0.5);

        }

        return mipmaps;

    }
}
