import { HttpRequest, HttpResponse } from "pi_pt/init/httprpc";
import { get_pid, to_pid } from "pi_pt/util/util";
import { initProjectDef } from "pi_pt/init/load_project";
import { AuthDbNew, AuthDbKey, UserDb, UserSessionDb } from "management_page/server/db/user_db.struct";
import { Env } from "pi_ucore/db/env";
import * as User from "./user.event";
import { Session } from "pi_ucore/event/session_new";
import { toByteArray } from "pi_tools/base64";
import { randomString } from "pi_tools/util";
import { checkAuth } from "./user.event";
import { AUTH, FM, RESULT_NUM } from "../../../constant";
import { getPlugin, pluginsList } from "../core/plugins_mgr";
import { sendRpcPid } from "../core/pid_event";
import { resultErr, resultOk } from "../result_handle";
import { Tr } from "pi_ucore/db/mgr";
import { getAllGrayVersions } from "../core/gray_lib";
import { getLogger } from 'pi_tools/logger';

declare var module;
declare var env: Env;
declare var readFile: any;

// #[httpRpc=/mgr/api/porxy/login]
export const login = async (req: HttpRequest, resp: HttpResponse) => {
    let msg = req.getData();
    let data = new UserDb();
    data.user = msg.user;
    data.pwd = msg.pwd;
    let localEnv = await getEnv(msg.user);
    let r = await User.login(data, localEnv);
    console.log("login!!!!!!!!!!! r:", r);
    if (r.status === "success") {
        let token = randomString(6);
        // 设置会话信息
        await env.dbMgr.write(async (tr: Tr) => {
            await SessionMgr.setSession(tr, msg.user, { user: msg.user, session: localEnv.get("session"), token: { token, create_time: Date.now() } })
        })
        r.data.token = token;
        r.data.user = msg.user;
        return r;
    }
    return r;
}

// http授权转mqtt授权
// #[rpc=rpcServer, src=*]
export const loginMqtt = async (authorization: string, env: Env) => {
    const userData = await checkHttpPortAuthHandle(authorization);
    // 替换session
    env.set("session", userData.session);
    return "success";

}

// #[httpRpc=/mgr/api/porxy/createUser]
export const createUser = async (req: HttpRequest, resp: HttpResponse) => {
    let userData = await checkHttpPortAuth(req);
    let msg = req.getData();
    let localEnv = await getEnv(userData.user);
    let data = new UserDb();
    data.user = msg.user;
    data.pwd = msg.pwd;
    return await User.createUser(data, localEnv);
}

// #[httpRpc=/mgr/api/porxy/delUser]
export const delUser = async (req: HttpRequest, resp: HttpResponse) => {
    let userData = await checkHttpPortAuth(req);
    let msg = req.getData();
    let localEnv = await getEnv(userData.user);
    return await User.delUser(msg.user, localEnv);
}

// #[httpRpc=/mgr/api/porxy/mdfPwd]
export const mdfPwd = async (req: HttpRequest, resp: HttpResponse) => {
    let userData = await checkHttpPortAuth(req);
    let msg = req.getData();
    let localEnv = await getEnv(userData.user);
    let data = new UserDb();
    data.user = msg.user;
    data.pwd = msg.pwd;
    return await User.mdfPwd(data, localEnv);
}

// #[httpRpc=/mgr/api/porxy/users]
export const users = async (req: HttpRequest, resp: HttpResponse) => {
    let userData = await checkHttpPortAuth(req);
    let localEnv = await getEnv(userData.user);
    return await User.users(undefined, localEnv);
}

// #[httpRpc=/mgr/api/porxy/setAuth]
export const setAuth = async (req: HttpRequest, resp: HttpResponse) => {
    let userData = await checkHttpPortAuth(req);
    let msg = req.getData();
    let json = JSON.parse(msg[""]);

    console.log("setAuth!!!!!!!!!!!!", json);
    let localEnv = await getEnv(userData.user);
    let auth = new AuthDbNew();
    auth.user = json.user;
    auth.auth = json.auth;
    return await User.setAuth(auth, localEnv);
}

// 管理端插件代理 body: {event: string, data: any}
// #[httpRpc=/mgr/api/porxy/plugins_rpc]
export const plugins = async (req: HttpRequest, resp: HttpResponse) => {
    let userData = await checkHttpPortAuth(req);
    let msg = req.getData();
    let json = JSON.parse(msg[""]);
    const event = json.event;
    const data = json.data;
    const timeout = json.timeout;
    const gray = json.gray;
    let localEnv = await getEnv(userData.user);

    const pluginData = await getPlugin(event);
    if (!pluginData) throw "插件不存在";
    logI("management plugins rpc ", { plugin_name: pluginData.plugin_name, user: userData.user, event, data, gray, timeout })
    // 权限字符串R、W、RW转换为AUTH枚举
    switch (pluginData.auth.toLowerCase()) {
        case "r":
            if (!checkAuth(pluginData.plugin_name, AUTH.R, localEnv)) throw "权限错误";
            break;
        case "w":
            if (!checkAuth(pluginData.plugin_name, AUTH.W, localEnv)) throw "权限错误";
            break;
        case "rw":
            if (!checkAuth(pluginData.plugin_name, AUTH.RW, localEnv)) throw "权限错误";
            break;
        default:
            throw "权限为空";
    }
    try {
        return resultOk(await sendRpcPid(event, data, timeout, gray));
    } catch (error) {
        console.log("plugins error:", error.message, error, error.stack);
        return resultErr(error.message);
    }
}

// 获取所有插件模块
// #[httpRpc=/mgr/api/porxy/get_plugin_names]
export const getPluginNames = async (req: HttpRequest, resp: HttpResponse) => {
    return pluginsList();
}


// 获取所有灰度版本
// #[httpRpc=/mgr/api/porxy/get_all_gray_versions]
export const getAllGrayVersionsPort = async (req: HttpRequest, resp: HttpResponse) => {
    return getAllGrayVersions();
}

// 获取灰度版本对应打开的端口
// #[httpRpc=/mgr/api/porxy/get_gray_versions_port]
export const getGrayVersionsPort = async (req: HttpRequest, resp: HttpResponse) => {
    const msg = req.getData();
    let path = msg.path;
    let data: any = readFile(path + "/ptconfig.json", false);
    data = JSON.parse(data);
    let r = {
        mqtt: [],
        http: []
    };
    for (let cfg of data.mqtt) {
        r.mqtt.push(cfg.port);
    }

    for (let cfg of data.httpConfig) {
        r.http.push(cfg.port);
    }

    return r;
}

// 获取env
export const getEnv = async (user: string): Promise<Env> => {
    return await env.dbMgr.read(async (tr: Tr) => {
        let localEnv = new Env();
        localEnv.dbMgr = env.dbMgr;
        localEnv.other = new Map();
        const userData = await SessionMgr.getSession(tr, user)
        let session = userData ? userData.session : new Session(Date.now().toString(), new Map(), () => { });

        localEnv.set("session", session);
        return localEnv;
    })

}


const mapToJson = (map: Map<string, any>) => {
    let json = {};
    map.forEach((v, k) => {
        json[k] = v;
    })
    return json;
}

const jsonToMap = (json: object) => {
    let map = new Map();
    for (let k of Object.keys(json)) {
        console.log("jsonToMap:", { k, v: json[k] });
        map.set(k, json[k]);
    }
    return map;
}

const setCookie = (resp: HttpResponse, cookie: string) => {
    resp.setHeader("Set-Cookie", `attr=${cookie}`);
}

const getCookie = (req: HttpRequest): string => {
    let cookie = req.getHeader("cookie");
    return cookie
}

type UserData = {
    user: string,
    session: Session,
    token: {
        token: string,
        create_time: number
    },
}

class SessionMgr {
    static async setSession(tr: Tr, user: string, userData: UserData) {
        let userSession = new UserSessionDb();
        userSession.user = user;
        userSession.session = userData.session.map;
        userSession.token = userData.token.token;
        userSession.create_time = userData.token.create_time;

        await tr.modify([{ ware: "memory", tab: UserSessionDb._$info.name, key: userSession.user, value: userSession }], 1000, true);
    }

    static async getSession(tr: Tr, user: string) {
        let userData: UserData;
        const r = (await tr.query([{ ware: "memory", tab: UserSessionDb._$info.name, key: user }], 1000, false))[0].value;
        if (r) {
            let session = new Session("1", r.session);
            userData = { user: user, session, token: { token: r.token, create_time: r.create_time } };
        }
        return userData;
    }
}

// httpPort权限检测
export const checkHttpPortAuth = async (req: HttpRequest): Promise<UserData> => {
    let auth = req.getHeader("authorization");
    // Basic dGVzdDoxMjM0NTY=
    return checkHttpPortAuthHandle(auth);
}

// httpPort权限检测
const checkHttpPortAuthHandle = async (authorization: string): Promise<UserData> => {
    try {
        let txt = new TextDecoder().decode(toByteArray(authorization.split(' ')[1]));
        const [username, password] = txt.split(':');
        let userData;
        await env.dbMgr.read(async (tr: Tr) => {
            userData = await SessionMgr.getSession(tr, username);
        })

        if (!userData) throw new Error("no user");
        if (password !== userData.token.token || userData.token.create_time + 7 * 24 * 60 * 60 * 1000 < Date.now()) {
            throw new Error("check auth error");
        };
        return userData;
    } catch (error) {
        console.log("checkHttpPortAuth auth error:", error.message, error.stack);
        throw new Error("auth error");
    }
}

const { logI, logD, logE, logW, logV } = getLogger(module.name);