/**
 * 用户管理
 */

import { AuthDbNew, AuthDbKey, UserDb } from "management_page/server/db/user_db.struct";
import { Env } from "pi_ucore/db/env";
import { Log } from "pi_pt/util/log";
import { AUTH, FM, RESULT_NUM } from "../../../constant";
import { Tr } from "pi_ucore/db/mgr";
import { User, Users } from "management_page/server/rpc/user.struct";
import { PT_ENV } from "pi_pt/constant";
import { Session } from "pi_ucore/event/session_new";
import { Ok } from "pi_net/rpc_r.struct";
import { resultErr, resultOk } from "../result_handle";

declare var env: Env;
// 登录
// #[rpc=rpcServer, src=*]
export const login = async (user: UserDb, env: Env) => {
    let session = <Session>env.get(PT_ENV.session);
    let userDb;
    await env.dbMgr.read(async (tr) => {
        userDb = <UserDb>(await tr.query([{ ware: "logfile", tab: UserDb._$info.name, key: user.user }], 1000, false))[0].value;
    });
    if (userDb && userDb.pwd === user.pwd) {
        let auth = await getAuths(user.user);
        // 写入会话表
        session.set("user", user.user);
        session.set("auth", auth as any);
        Log.info("login ok ", { user: user.user });
        console.log("auth!!!!!!!!!", auth);
        return resultOk(auth);
    } else if (!userDb && user.user === 'admin') {
        // admin特殊处理，不存在则初始化
        session.set("user", user.user);
        await createUser(user, env);
        let auth = await getAuths(user.user);
        session.set("auth", auth as any);
        return resultOk(auth);
    } else {
        return resultErr('账号或密码错误');
    }
};

// 创建用户
// #[rpc=rpcServer, src=*]
export const createUser = async (user: UserDb, env: Env) => {
    if (!checkAuth(FM.USER_MGR, AUTH.W, env)) throw "权限错误";
    return await env.dbMgr.write(async (tr) => {
        let userDb = <UserDb>(await tr.query([{ ware: "logfile", tab: UserDb._$info.name, key: user.user }], 1000, false))[0].value;
        if (!userDb) {
            await tr.modify([{ ware: "logfile", tab: UserDb._$info.name, key: user.user, value: user }], 1000, true);
            return resultOk();
        } else {
            return resultErr("用户已存在");
        }
    });
}

// 删除用户
// #[rpc=rpcServer, src=*]
export const delUser = async (user: string, env: Env) => {
    let session = <Session>env.get(PT_ENV.session);
    let sessionUser = session.get("user");
    if (sessionUser === user) return resultErr("不能删除自己");
    if (!checkAuth(FM.USER_MGR, AUTH.W, env)) return resultErr("权限错误");
    return await env.dbMgr.write(async (tr) => {
        // 判断用户是否存在
        let userDb = <UserDb>(await tr.query([{ ware: "logfile", tab: UserDb._$info.name, key: user }], 1000, false))[0].value;

        if (!userDb) { return resultErr("用户不存在") };
        await tr.modify([
            { ware: "logfile", tab: AuthDbNew._$info.name, key: user },
            { ware: "logfile", tab: UserDb._$info.name, key: user }
        ], 1000, true);
        Log.info("del user ok ", { user });
        return resultOk();
    });

}

// 修改密码
// #[rpc=rpcServer, src=*]
export const mdfPwd = async (user: UserDb, env: Env) => {
    // 判断是否修改自己的密码
    let session = <Session>env.get(PT_ENV.session);
    let sessionUser = session.get("user");
    if (!checkAuth(FM.USER_MGR, AUTH.W, env) && sessionUser !== user.user) return resultErr("权限错误");
    return await env.dbMgr.write(async (tr) => {
        let userDb = <UserDb>(await tr.query([{ ware: "logfile", tab: UserDb._$info.name, key: user.user }], 1000, false))[0].value;
        if (userDb) {
            await tr.modify([{ ware: "logfile", tab: UserDb._$info.name, key: user.user, value: user }], 1000, true);
            Log.info("modification pwd ok ", { user: user.user });
            return resultOk();
        } else {
            return resultErr('用户不存在');
        }
    });
}

// 获取所有用户
// #[rpc=rpcServer, src=*]
export const users = async (_arg, env: Env) => {
    if (!checkAuth(FM.USER_MGR, AUTH.R, env)) return resultErr("权限错误");
    let users = new Users();
    users.list = [];
    await env.dbMgr.read(async (tr) => {
        let iter = tr.kvList("logfile", UserDb._$info.name, undefined, false, '');
        for await (const item of iter) {
            let user = item[1] as UserDb;
            users.list.push(await getAuths(user.user, tr));
        }
    })
    Log.info("get users ok ");
    return resultOk(users);
}

// 设置用户权限
// #[rpc=rpcServer, src=*]
export const setAuth = async (auth: AuthDbNew, env: Env) => {
    console.log("auth:", JSON.stringify(auth));
    if (!checkAuth(FM.USER_MGR, AUTH.W, env)) return resultErr("权限错误");
    await env.dbMgr.write(async (tr) => {
        await tr.modify([{ ware: "logfile", tab: AuthDbNew._$info.name, key: auth.user, value: auth }], 1000, true);
    });
    Log.info("set auth ok ", { auth });

    return resultOk();
}

// 获取用户权限
export const getAuths = async (user: string, tr?: Tr): Promise<AuthDbNew> => {
    let func = async (tr: Tr) => {
        let auth = <AuthDbNew>(await tr.query([{ ware: "logfile", tab: AuthDbNew._$info.name, key: user }], 1000, false))[0].value;
        if (!auth) {
            auth = new AuthDbNew();
            auth.user = user;
            auth.auth = {};
        }
        return auth;
    }

    if (!tr) {
        return await env.dbMgr.read(async tr => {
            return await func(tr);
        })
    } else {
        return await func(tr);
    }
}

// 检测权限配置
export const checkAuth = (module: FM | string, id: AUTH, env: Env) => {
    // return true;
    let session = <Session>env.get(PT_ENV.session);
    if (session.get("user") === "admin") return true;

    let auth = <any>session.get("auth") as AuthDbNew;
    if (!auth) return false;
    let authID = auth.auth[module];
    if ((authID & id) === id) {
        return true;
    }
    return false
}