import type { UserInfo } from '/#/store'
import type { ErrorMessageMode } from '/#/axios'
import { defineStore } from 'pinia'
import { store } from '/@/store'
import { RoleEnum } from '/@/enums/roleEnum'
import { PageEnum } from '/@/enums/pageEnum'
import { ROLES_KEY, TOKEN_KEY, USER_INFO_KEY } from '/@/enums/cacheEnum'
import { getAuthCache, setAuthCache } from '/@/utils/auth'
import { GetUserInfoModel, LoginParams } from '/@/api/sys/model/userModel'
import { doLogout, loginApi } from '/@/api/sys/user'
import { getUserInfo } from '/@/api/sys/userLogin'
import { useI18n } from '/@/hooks/web/useI18n'
import { ConfirmOptions, useMessage } from '/@/hooks/web/useMessage'
import { router } from '/@/router'
import { usePermissionStore } from '/@/store/modules/permission'
import { RouteRecordRaw } from 'vue-router'
import { PAGE_NOT_FOUND_ROUTE } from '/@/router/routes/basic'
import { isArray } from '/@/utils/is'
import { h } from 'vue'
import { message, Modal, notification } from 'ant-design-vue'


interface UserState {
    userInfo: Nullable<UserInfo>
    token?: string
    roleList: RoleEnum[]
    sessionTimeout?: boolean
    lastUpdateTime: number
    permissionHide: any
}

export const useUserStore = defineStore({
    id: 'app-user',
    state: (): UserState => ({
        // user info
        userInfo: null,
        // token
        token: undefined,
        // roleList
        roleList: [],
        // Whether the login expired
        sessionTimeout: false,
        // Last fetch time
        lastUpdateTime: 0,
        // hide菜单
        permissionHide: {
            db_mgr: { path: '/DBManger', btnPower: false },
            user_mgr: { path: '/UserManger', btnPower: false },
            auth_mgr: { path: false, btnPower: false },
            web_shell: { path: '/webshell', btnPower: false },
            client_web_shell: { path: '/webshellfront', btnPower: false },
            welcome: { path: '/welcome', btnPower: false },
            pt_load_info: { path: '/VMManager', btnPower: false },
            rpc_console: { path: '/rpcshell', btnPower: false },
        },
    }),
    getters: {
        getUserInfo(): UserInfo {
            return this.userInfo || getAuthCache<UserInfo>(USER_INFO_KEY) || {}
        },
        getToken(): string {
            return this.token || getAuthCache<string>(TOKEN_KEY)
        },
        getRoleList(): RoleEnum[] {
            return this.roleList.length > 0 ? this.roleList : getAuthCache<RoleEnum[]>(ROLES_KEY)
        },
        getSessionTimeout(): boolean {
            return !!this.sessionTimeout
        },
        getLastUpdateTime(): number {
            return this.lastUpdateTime
        },
        getPermissionHide(): any {
            return this.permissionHide
        },
    },
    actions: {
        setToken(info: string | undefined) {
            this.token = info ? info : '' // for null or undefined value
            setAuthCache(TOKEN_KEY, info)
        },
        setRoleList(roleList: RoleEnum[]) {
            this.roleList = roleList
            setAuthCache(ROLES_KEY, roleList)
        },
        setUserInfo(info: UserInfo | null) {
            this.userInfo = info
            this.lastUpdateTime = new Date().getTime()
            setAuthCache(USER_INFO_KEY, info)
        },
        setSessionTimeout(flag: boolean) {
            this.sessionTimeout = flag
        },
        setPermissionHide(flag: any) {
            this.permissionHide = flag
        },
        resetState() {
            this.userInfo = null
            this.token = ''
            this.roleList = []
            this.sessionTimeout = false
        },
        /**
         * @description: login
         */
        // async login(
        //     params: LoginParams & {
        //         goHome?: boolean
        //         mode?: ErrorMessageMode
        //     },
        // ): Promise<GetUserInfoModel | null> {
        //     try {
        //         const { goHome = true, mode, ...loginParams } = params
        //         const data = await loginApi(loginParams, mode)
        //         const { token } = data

        //         // save token
        //         this.setToken(token)
        //         return this.afterLoginAction(goHome)
        //     } catch (error) {
        //         return Promise.reject(error)
        //     }
        // },
        async afterLoginAction(params?, goHome?: boolean): Promise<UserInfo | null | any> {
            if (!this.getToken) return null
            // get user info
            const userInfo = await this.getUserInfoAction(params)
            const changePower = () => {
                let newUserPower: any = {
                    db_mgr: { path: null, btnPower: false },
                    user_mgr: { path: null, btnPower: false },
                    auth_mgr: { path: false, btnPower: false },
                    web_shell: { path: null, btnPower: false },
                    client_web_shell: { path: null, btnPower: false },
                    welcome: { path: null, btnPower: false },
                    pt_load_info: { path: null, btnPower: false },
                    rpc_console: { path: null, btnPower: false },
                }
                let userPower = this.getUserInfo.auth;
                if (userPower.hasOwnProperty('db_mgr')) {
                    const result = userPower.db_mgr;
                    if (result >= 2) {
                        newUserPower.db_mgr.btnPower = true;
                        newUserPower.db_mgr.path = '/DBManger';
                    } else {
                        newUserPower.db_mgr.path = '/DBManger';
                    }
                }
                if (userPower.hasOwnProperty('user_mgr')) {
                    const result = userPower.user_mgr;
                    if (result >= 2) {
                        newUserPower.user_mgr.btnPower = true;
                        newUserPower.user_mgr.path = '/UserManger';
                    } else {
                        newUserPower.user_mgr.path = '/UserManger';
                    }
                }
                if (userPower.hasOwnProperty('auth_mgr')) {
                    const result = userPower.auth_mgr;
                    if (result >= 2) {
                        newUserPower.auth_mgr.btnPower = true;
                        newUserPower.auth_mgr.path = true;
                    } else {
                        newUserPower.auth_mgr.path = true;
                    }
                }
                if (userPower.hasOwnProperty('web_shell')) {
                    const result = userPower.web_shell;
                    if (result >= 2) {
                        newUserPower.web_shell.btnPower = true;
                        newUserPower.web_shell.path = '/webshell';
                    }
                }
                if (userPower.hasOwnProperty('client_web_shell')) {
                    const result = userPower.client_web_shell;
                    if (result >= 2) {
                        newUserPower.client_web_shell.btnPower = true;
                        newUserPower.client_web_shell.path = '/webshellfront';
                    } else {
                        newUserPower.client_web_shell.path = '/webshellfront';
                    }
                }
                if (userPower.hasOwnProperty('pt_load_info')) {
                    const result = userPower.pt_load_info;
                    if (result >= 2) {
                        newUserPower.pt_load_info.btnPower = true;
                        newUserPower.pt_load_info.path = '/VMManager';
                    } else {
                        newUserPower.pt_load_info.path = '/VMManager';
                    }
                }
                if (userPower.hasOwnProperty('rpc_console')) {
                    const result = userPower.rpc_console;
                    if (result >= 2) {
                        newUserPower.rpc_console.btnPower = true;
                        newUserPower.rpc_console.path = '/rpcshell';
                    } else {
                        newUserPower.rpc_console.path = '/rpcshell';
                    }
                }
                if (this.getUserInfo.user !== 'admin') {
                    if (Object.keys(userPower).length !== 0) {
                        newUserPower.welcome.path = '/welcome';
                    }
                }
                return newUserPower;
            }
            this.setPermissionHide(changePower());
            const sessionTimeout = this.sessionTimeout
            if (sessionTimeout) {
                this.setSessionTimeout(false)
            } else {
                const permissionStore = usePermissionStore()
                if (!permissionStore.isDynamicAddedRoute) {
                    const routes = await permissionStore.buildRoutesAction();
                    let power = changePower();
                    const objectValues = Object.values(power);
                    routes.forEach((route) => {
                        const hasTarget = objectValues.includes(route.name) && userInfo?.user !== 'admin';
                        if (hasTarget) return;
                        router.addRoute(route as unknown as RouteRecordRaw)
                    })
                    router.addRoute(PAGE_NOT_FOUND_ROUTE as unknown as RouteRecordRaw);

                    permissionStore.setDynamicAddedRoute(true)
                }

                goHome && (await router.replace(userInfo?.homePath || PageEnum.BASE_HOME))
            }
            return userInfo
        },
        async getUserInfoAction(params?): Promise<UserInfo | null> {
            if (!this.getToken) return null
            if (!params) {
                router.push(PageEnum.BASE_LOGIN);
            }
            const userInfo = params
            const { roles = [] } = userInfo
            if (isArray(roles)) {
                const roleList = roles.map((item) => item.value) as RoleEnum[]
                this.setRoleList(roleList)
            } else {
                userInfo.roles = []
                this.setRoleList([])
            }
            this.setUserInfo(userInfo)
            return userInfo
        },
        /**
         * @description: logout
         */
        async logout(goLogin = false) {
            if (!this.getToken) {
                // try {
                //     await doLogout()
                // } catch {
                //     console.log('注销Token失败')
                // }
                return
            }
            this.setToken(undefined)
            this.setSessionTimeout(false)
            this.setUserInfo(null)
            goLogin && router.push(PageEnum.BASE_LOGIN)
        },

        /**
         * @description: Confirm before logging out
         */
        confirmLoginOut() {
            const { createConfirm } = useMessage()
            const { t } = useI18n()
            createConfirm({
                iconType: 'warning',
                title: () => h('span', t('sys.app.Tip')),
                content: () => h('span', t('sys.app.logoutMessage')),
                onOk: async () => {
                    await this.logout(true);
                },
                onCancel: () => {
                }
            })
        },

    },
})

// Need to be used outside the setup
export function useUserStoreWithOut() {
    return useUserStore(store)
}
